grammar TrinoSqlParser;

tokens {
	DELIMITER
}

// Modified entrypoint
program: statements* EOF;

statements:
	singleStatement
	| standaloneExpression
	| standalonePathSpecification
	| standaloneType
	| standaloneRowPattern;

singleStatement: statement SEMICOLON?;

standaloneExpression: expression SEMICOLON?;

standalonePathSpecification: pathSpecification SEMICOLON?;

standaloneType: type SEMICOLON?;

standaloneRowPattern: rowPattern SEMICOLON?;

statement:
	query												# statementDefault
	| USE schema = identifier							# use
	| USE catalog = identifier '.' schema = identifier	# use
	| CREATE SCHEMA (IF NOT EXISTS)? qualifiedName (
		AUTHORIZATION principal
	)? (WITH properties)?											# createSchema
	| DROP SCHEMA (IF EXISTS)? qualifiedName (CASCADE | RESTRICT)?	# dropSchema
	| ALTER SCHEMA qualifiedName RENAME TO identifier				# renameSchema
	| ALTER SCHEMA qualifiedName SET AUTHORIZATION principal		# setSchemaAuthorization
	| CREATE TABLE (IF NOT EXISTS)? qualifiedName columnAliases? (
		COMMENT string
	)? (WITH properties)? AS (query | '(' query ')') (
		WITH (NO)? DATA
	)? # createTableAsSelect
	| CREATE TABLE (IF NOT EXISTS)? qualifiedName '(' tableElement (
		',' tableElement
	)* ')' (COMMENT string)? (WITH properties)?										# createTable
	| DROP TABLE (IF EXISTS)? qualifiedName											# dropTable
	| INSERT INTO qualifiedName columnAliases? query								# insertInto
	| DELETE FROM qualifiedName (WHERE booleanExpression)?	# delete
	| TRUNCATE TABLE qualifiedName						# truncateTable 
	| ALTER TABLE (IF EXISTS)? from = qualifiedName RENAME TO to = qualifiedName	# renameTable
	| COMMENT ON TABLE qualifiedName IS (string | NULL)								# commentTable
	| COMMENT ON COLUMN qualifiedName IS (string | NULL)							# commentColumn
	| ALTER TABLE (IF EXISTS)? tableName = qualifiedName RENAME COLUMN (
		IF EXISTS
	)? from = identifier TO to = identifier # renameColumn
	| ALTER TABLE (IF EXISTS)? tableName = qualifiedName DROP COLUMN (
		IF EXISTS
	)? column = qualifiedName # dropColumn
	| ALTER TABLE (IF EXISTS)? tableName = qualifiedName ADD COLUMN (
		IF NOT EXISTS
	)? column = columnDefinition										# addColumn
	| ALTER TABLE tableName = qualifiedName SET AUTHORIZATION principal				# setTableAuthorization
	| ALTER TABLE tableName = qualifiedName SET PROPERTIES propertyAssignments	# setTableProperties
	| ALTER TABLE tableName = qualifiedName EXECUTE procedureName = identifier (
		'(' (callArgument (',' callArgument)*)? ')'
	)? (WHERE where = booleanExpression)? # tableExecute
	| ANALYZE qualifiedName (WITH properties)?							# analyze
	| CREATE (OR REPLACE)? MATERIALIZED VIEW (IF NOT EXISTS)? qualifiedName (
		COMMENT string
	)? (WITH properties)? AS query # createMaterializedView
	| CREATE (OR REPLACE)? VIEW qualifiedName (COMMENT string)? (
		SECURITY (DEFINER | INVOKER)
	)? AS query															# createView
	| REFRESH MATERIALIZED VIEW qualifiedName							# refreshMaterializedView
	| DROP MATERIALIZED VIEW (IF EXISTS)? qualifiedName											# dropMaterializedView
	| ALTER MATERIALIZED VIEW (IF EXISTS)? from = qualifiedName RENAME TO to = qualifiedName	# renameMaterializedView
	| ALTER MATERIALIZED VIEW qualifiedName 
		SET PROPERTIES propertyAssignments # setMaterializedViewProperties
	| DROP VIEW (IF EXISTS)? qualifiedName								# dropView
	| ALTER VIEW from = qualifiedName RENAME TO to = qualifiedName		# renameView
	| ALTER VIEW from = qualifiedName SET AUTHORIZATION principal		# setViewAuthorization
	| CALL qualifiedName '(' (callArgument (',' callArgument)*)? ')'	# call
	| CREATE ROLE name = identifier 
		(WITH ADMIN grantor)? 
		(IN catalog = identifier)?										# createRole 
	| DROP ROLE name = identifier										# dropRole
	| GRANT roles TO principal 
	(',' principal)*
	(WITH ADMIN OPTION)? 
	(GRANTED BY grantor)? 
	(IN catalog = identifier)?                                          # grantRoles
	| REVOKE (ADMIN OPTION FOR)? roles FROM principal 
	(',' principal)* 
	(GRANTED BY grantor)?
	(IN catalog = identifier)? # revokeRoles
	| SET ROLE (ALL | NONE | role = identifier) (
		IN catalog = identifier
	)? # setRole | GRANT (privilege (',' privilege)* | ALL PRIVILEGES) ON (
		SCHEMA
		| TABLE
	)? qualifiedName TO grantee = principal (WITH GRANT OPTION)? # grant
	| DENY (privilege (',' privilege)* | ALL PRIVILEGES) ON (
		SCHEMA
		| TABLE
	)? qualifiedName TO grantee = principal # deny
	| REVOKE (GRANT OPTION FOR)? (
		privilege (',' privilege)*
		| ALL PRIVILEGES
	) ON (SCHEMA | TABLE)? qualifiedName FROM grantee = principal	# revoke
	| SHOW GRANTS (ON TABLE? qualifiedName)?						# showGrants
	| EXPLAIN ANALYZE? VERBOSE? (
		'(' explainOption (',' explainOption)* ')'
	)? statement									# explain
	| SHOW CREATE TABLE qualifiedName				# showCreateTable
	| SHOW CREATE SCHEMA qualifiedName				# showCreateSchema
	| SHOW CREATE VIEW qualifiedName				# showCreateView
	| SHOW CREATE MATERIALIZED VIEW qualifiedName	# showCreateMaterializedView
	| SHOW TABLES ((FROM | IN) qualifiedName)? (
		LIKE pattern = string (ESCAPE escape = string)?
	)? # showTables
	| SHOW SCHEMAS ((FROM | IN) identifier)? (
		LIKE pattern = string (ESCAPE escape = string)?
	)? # showSchemas
	| SHOW CATALOGS (
		LIKE pattern = string (ESCAPE escape = string)?
	)? # showCatalogs
	| SHOW COLUMNS (FROM | IN) qualifiedName? (
		LIKE pattern = string (ESCAPE escape = string)?
	)?												# showColumns
	| SHOW STATS FOR qualifiedName					# showStats
	| SHOW STATS FOR '(' query ')'					# showStatsForQuery
	| SHOW CURRENT? ROLES ((FROM | IN) identifier)?	# showRoles
	| SHOW ROLE GRANTS ((FROM | IN) identifier)?	# showRoleGrants
	| DESCRIBE qualifiedName						# showColumns
	| DESC qualifiedName							# showColumns
	| SHOW FUNCTIONS (
		LIKE pattern = string (ESCAPE escape = string)?
	)? # showFunctions
	| SHOW SESSION (
		LIKE pattern = string (ESCAPE escape = string)?
	)?																# showSession
	| SET SESSION qualifiedName EQ expression						# setSession
	| RESET SESSION qualifiedName									# resetSession
	| START TRANSACTION (transactionMode (',' transactionMode)*)?	# startTransaction
	| COMMIT WORK?													# commit
	| ROLLBACK WORK?												# rollback
	| PREPARE identifier FROM statement								# prepare
	| DEALLOCATE PREPARE identifier									# deallocate
	| EXECUTE identifier (USING expression (',' expression)*)?		# execute
	| DESCRIBE INPUT identifier										# describeInput
	| DESCRIBE OUTPUT identifier									# describeOutput
	| SET PATH pathSpecification									# setPath
	| SET TIME ZONE (LOCAL | expression)							# setTimeZone
	| UPDATE qualifiedName SET updateAssignment (
		',' updateAssignment
	)* (WHERE where = booleanExpression)?													# update
	| MERGE INTO qualifiedName (AS? identifier)? USING relation ON expression mergeCase+	# merge
	| SHOW COMMENT ON TABLE qualifiedName													# showTableComment
	| SHOW COMMENT ON COLUMN qualifiedName													# showColumnComment;

query: with? queryNoWith;

with: WITH RECURSIVE? namedQuery (',' namedQuery)*;

tableElement: columnDefinition | likeClause;

columnDefinition:
	identifier type (NOT NULL)? (COMMENT string)? (
		WITH properties
	)?;

likeClause:
	LIKE qualifiedName (
		optionType = (INCLUDING | EXCLUDING) PROPERTIES
	)?;

properties: '(' propertyAssignments ')';

propertyAssignments: property (',' property)*;

property: identifier EQ propertyValue;

propertyValue:
	DEFAULT			# defaultPropertyValue
	| expression	# nonDefaultPropertyValue;

queryNoWith:
	queryTerm (ORDER BY sortItem (',' sortItem)*)? (
		OFFSET offset = rowCount (ROW | ROWS)?
	)? (
		(LIMIT limit = limitRowCount)
		| (
			FETCH (FIRST | NEXT) (fetchFirst = rowCount)? (
				ROW
				| ROWS
			) (ONLY | WITH TIES)
		)
	)?;

limitRowCount: ALL | rowCount;

rowCount: INTEGER_VALUE | QUESTION_MARK;

queryTerm:
	queryPrimary																	# queryTermDefault
	| left = queryTerm operator = INTERSECT setQuantifier? right = queryTerm		# setOperation
	| left = queryTerm operator = (UNION | EXCEPT) setQuantifier? right = queryTerm	# setOperation;

queryPrimary:
	querySpecification						# queryPrimaryDefault
	| TABLE qualifiedName					# table
	| VALUES expression (',' expression)*	# inlineTable
	| '(' queryNoWith ')'					# subquery;

sortItem:
	expression ordering = (ASC | DESC)? (
		NULLS nullOrdering = (FIRST | LAST)
	)?;

querySpecification:
	SELECT setQuantifier? selectItem (',' selectItem)* (
		FROM relation (',' relation)*
	)? (WHERE where = booleanExpression)? (GROUP BY groupBy)? (
		HAVING having = booleanExpression
	)? (WINDOW windowDefinition (',' windowDefinition)*)?;

groupBy: setQuantifier? groupingElement (',' groupingElement)*;

groupingElement:
	groupingSet												# singleGroupingSet
	| ROLLUP '(' (expression (',' expression)*)? ')'		# rollup
	| CUBE '(' (expression (',' expression)*)? ')'			# cube
	| GROUPING SETS '(' groupingSet (',' groupingSet)* ')'	# multipleGroupingSets;

groupingSet:
	'(' (expression (',' expression)*)? ')'
	| expression;

windowDefinition:
	name = identifier AS '(' windowSpecification ')';

windowSpecification: (existingWindowName = identifier)? (
		PARTITION BY partition += expression (
			',' partition += expression
		)*
	)? (ORDER BY sortItem (',' sortItem)*)? windowFrame?;

namedQuery: name = identifier (columnAliases)? AS '(' query ')';

setQuantifier: DISTINCT | ALL;

selectItem:
	expression (AS? identifier)?							# selectSingle
	| primaryExpression '.' ASTERISK (AS columnAliases)?	# selectAll
	| ASTERISK												# selectAll;

relation:
	left = relation (
		CROSS JOIN right = sampledRelation
		| joinType JOIN rightRelation = relation joinCriteria
		| NATURAL joinType JOIN right = sampledRelation
	)					# joinRelation
	| sampledRelation	# relationDefault;

joinType: INNER? | LEFT OUTER? | RIGHT OUTER? | FULL OUTER?;

joinCriteria:
	ON booleanExpression
	| USING '(' identifier (',' identifier)* ')';

sampledRelation:
	patternRecognition (
		TABLESAMPLE sampleType '(' percentage = expression ')'
	)?;

sampleType: BERNOULLI | SYSTEM;

patternRecognition:
	aliasedRelation (
		MATCH_RECOGNIZE '(' (
			PARTITION BY partition += expression (
				',' partition += expression
			)*
		)? (ORDER BY sortItem (',' sortItem)*)? (
			MEASURES measureDefinition (',' measureDefinition)*
		)? rowsPerMatch? (AFTER MATCH skipTo)? (INITIAL | SEEK)? PATTERN '(' rowPattern ')' (
			SUBSET subsetDefinition (',' subsetDefinition)*
		)? DEFINE variableDefinition (',' variableDefinition)* ')' (
			AS? identifier columnAliases?
		)?
	)?;

measureDefinition: expression AS identifier;

rowsPerMatch:
	ONE ROW PER MATCH
	| ALL ROWS PER MATCH emptyMatchHandling?;

emptyMatchHandling:
	SHOW EMPTY MATCHES
	| OMIT EMPTY MATCHES
	| WITH UNMATCHED ROWS;

skipTo:
	'SKIP' TO NEXT ROW
	| 'SKIP' PAST LAST ROW
	| 'SKIP' TO FIRST identifier
	| 'SKIP' TO LAST identifier
	| 'SKIP' TO identifier;

subsetDefinition:
	name = identifier EQ '(' union += identifier (
		',' union += identifier
	)* ')';

variableDefinition: identifier AS expression;

aliasedRelation:
	relationPrimary (AS? identifier columnAliases?)?;

columnAliases: '(' identifier (',' identifier)* ')';

relationPrimary:
	qualifiedName	# tableName
	| '(' query ')'	# subqueryRelation
	| UNNEST '(' expression (',' expression)* ')' (
		WITH ORDINALITY
	)?						# unnest
	| LATERAL '(' query ')'	# lateral
	| '(' relation ')'		# parenthesizedRelation;

expression: booleanExpression;

booleanExpression:
	valueExpression predicate[$valueExpression.ctx]?					# predicated
	| NOT booleanExpression												# logicalNot
	| left = booleanExpression operator = AND right = booleanExpression	# logicalBinary
	| left = booleanExpression operator = OR right = booleanExpression	# logicalBinary;

// workaround for https://github.com/antlr/antlr4/issues/780
predicate[ParserRuleContext value]:
	comparisonOperator right = valueExpression							# comparison
	| comparisonOperator comparisonQuantifier '(' query ')'				# quantifiedComparison
	| NOT? BETWEEN lower = valueExpression AND upper = valueExpression	# between
	| NOT? IN '(' expression (',' expression)* ')'						# inList
	| NOT? IN '(' query ')'												# inSubquery
	| NOT? LIKE pattern = valueExpression (
		ESCAPE escape = valueExpression
	)?												# like
	| IS NOT? NULL									# nullPredicate
	| IS NOT? DISTINCT FROM right = valueExpression	# distinctFrom;

valueExpression:
	primaryExpression							# valueExpressionDefault
	| valueExpression AT timeZoneSpecifier		# atTimeZone
	| operator = (MINUS | PLUS) valueExpression	# arithmeticUnary
	| left = valueExpression operator = (
		ASTERISK
		| SLASH
		| PERCENT
	) right = valueExpression													# arithmeticBinary
	| left = valueExpression operator = (PLUS | MINUS) right = valueExpression	# arithmeticBinary
	| left = valueExpression CONCAT right = valueExpression						# concatenation;

primaryExpression:
	NULL													# nullLiteral
	| interval												# intervalLiteral
	| identifier string										# typeConstructor
	| DOUBLE PRECISION string								# typeConstructor
	| number												# numericLiteral
	| booleanValue											# booleanLiteral
	| string												# stringLiteral
	| BINARY_LITERAL										# binaryLiteral
	| QUESTION_MARK											# parameter
	| POSITION '(' valueExpression IN valueExpression ')'	# position
	| '(' expression (',' expression)+ ')'					# rowConstructor
	| ROW '(' expression (',' expression)* ')'				# rowConstructor
	| qualifiedName '(' ASTERISK ')' filter? over?			# functionCall
	| processingMode? qualifiedName '(' (
		setQuantifier? expression (',' expression)*
	)? (ORDER BY sortItem (',' sortItem)*)? ')' filter? (
		nullTreatment? over
	)?															# functionCall
	| identifier over											# measure
	| identifier '->' expression								# lambda
	| '(' (identifier (',' identifier)*)? ')' '->' expression	# lambda
	| '(' query ')'												# subqueryExpression
	// This is an extension to ANSI SQL, which considers EXISTS to be a <boolean expression>
	| EXISTS '(' query ')' # exists
	| CASE operand = expression whenClause+ (
		ELSE elseExpression = expression
	)? END														# simpleCase
	| CASE whenClause+ (ELSE elseExpression = expression)? END	# searchedCase
	| CAST '(' expression AS type ')'							# cast
	| TRY_CAST '(' expression AS type ')'						# cast
	| ARRAY '[' (expression (',' expression)*)? ']'				# arrayConstructor
	| value = primaryExpression '[' index = valueExpression ']'	# subscript
	| identifier												# columnReference
	| base = primaryExpression '.' fieldName = identifier		# dereference
	| name = CURRENT_DATE										# specialDateTimeFunction
	| name = CURRENT_TIME ('(' precision = INTEGER_VALUE ')')?	# specialDateTimeFunction
	| name = CURRENT_TIMESTAMP (
		'(' precision = INTEGER_VALUE ')'
	)?																# specialDateTimeFunction
	| name = LOCALTIME ('(' precision = INTEGER_VALUE ')')?			# specialDateTimeFunction
	| name = LOCALTIMESTAMP ('(' precision = INTEGER_VALUE ')')?	# specialDateTimeFunction
	| name = CURRENT_USER											# currentUser
	| name = CURRENT_CATALOG										# currentCatalog
	| name = CURRENT_SCHEMA											# currentSchema
	| name = CURRENT_PATH											# currentPath
	| SUBSTRING '(' valueExpression FROM valueExpression (
		FOR valueExpression
	)? ')'														# substring
	| NORMALIZE '(' valueExpression (',' normalForm)? ')'		# normalize
	| EXTRACT '(' identifier FROM valueExpression ')'			# extract
	| '(' expression ')'										# parenthesizedExpression
	| GROUPING '(' (qualifiedName (',' qualifiedName)*)? ')'	# groupingOperation;

processingMode: RUNNING | FINAL;

nullTreatment: IGNORE NULLS | RESPECT NULLS;

string:
	STRING								# basicStringLiteral
	| UNICODE_STRING (UESCAPE STRING)?	# unicodeStringLiteral;

timeZoneSpecifier:
	TIME ZONE interval	# timeZoneInterval
	| TIME ZONE string	# timeZoneString;

comparisonOperator: EQ | NEQ | LT | LTE | GT | GTE;

comparisonQuantifier: ALL | SOME | ANY;

booleanValue: TRUE | FALSE;

interval:
	INTERVAL sign = (PLUS | MINUS)? string from = intervalField (
		TO to = intervalField
	)?;

intervalField: YEAR | MONTH | DAY | HOUR | MINUTE | SECOND;

normalForm: NFD | NFC | NFKD | NFKC;

type:
	ROW '(' rowField (',' rowField)* ')'						# rowType
	| INTERVAL from = intervalField (TO to = intervalField)?	# intervalType
	| base = TIMESTAMP ('(' precision = typeParameter ')')? (
		WITHOUT TIME ZONE
	)?																		# dateTimeType
	| base = TIMESTAMP ('(' precision = typeParameter ')')? WITH TIME ZONE	# dateTimeType
	| base = TIME ('(' precision = typeParameter ')')? (
		WITHOUT TIME ZONE
	)?																	# dateTimeType
	| base = TIME ('(' precision = typeParameter ')')? WITH TIME ZONE	# dateTimeType
	| DOUBLE PRECISION													# doublePrecisionType
	| ARRAY '<' type '>'												# legacyArrayType
	| MAP '<' keyType = type ',' valueType = type '>'					# legacyMapType
	| type ARRAY ('[' INTEGER_VALUE ']')?								# arrayType
	| identifier ('(' typeParameter (',' typeParameter)* ')')?			# genericType;

rowField: type | identifier type;

typeParameter: INTEGER_VALUE | type;

whenClause:
	WHEN condition = expression THEN result = expression;

filter: FILTER '(' WHERE booleanExpression ')';

mergeCase:
	WHEN MATCHED (AND condition = expression)? THEN UPDATE SET targets += identifier EQ values +=
		expression (
		',' targets += identifier EQ values += expression
	)*															# mergeUpdate
	| WHEN MATCHED (AND condition = expression)? THEN DELETE	# mergeDelete
	| WHEN NOT MATCHED (AND condition = expression)? THEN INSERT (
		'(' targets += identifier (',' targets += identifier)* ')'
	)? VALUES '(' values += expression (',' values += expression)* ')' # mergeInsert;

over:
	OVER (windowName = identifier | '(' windowSpecification ')');

windowFrame: (
		MEASURES measureDefinition (',' measureDefinition)*
	)? frameExtent (AFTER MATCH skipTo)? (INITIAL | SEEK)? (
		PATTERN '(' rowPattern ')'
	)? (SUBSET subsetDefinition (',' subsetDefinition)*)? (
		DEFINE variableDefinition (',' variableDefinition)*
	)?;

frameExtent:
	frameType = RANGE start = frameBound
	| frameType = ROWS start = frameBound
	| frameType = GROUPS start = frameBound
	| frameType = RANGE BETWEEN start = frameBound AND end = frameBound
	| frameType = ROWS BETWEEN start = frameBound AND end = frameBound
	| frameType = GROUPS BETWEEN start = frameBound AND end = frameBound;

frameBound:
	UNBOUNDED boundType = PRECEDING						# unboundedFrame
	| UNBOUNDED boundType = FOLLOWING					# unboundedFrame
	| CURRENT ROW										# currentRowBound
	| expression boundType = (PRECEDING | FOLLOWING)	# boundedFrame;

rowPattern:
	patternPrimary patternQuantifier?	# quantifiedPrimary
	| rowPattern rowPattern				# patternConcatenation
	| rowPattern '|' rowPattern			# patternAlternation;

patternPrimary:
	identifier										# patternVariable
	| '(' ')'										# emptyPattern
	| PERMUTE '(' rowPattern (',' rowPattern)* ')'	# patternPermutation
	| '(' rowPattern ')'							# groupedPattern
	| '^'											# partitionStartAnchor
	| '$'											# partitionEndAnchor
	| '{-' rowPattern '-}'							# excludedPattern;

patternQuantifier:
	ASTERISK (reluctant = QUESTION_MARK)?							# zeroOrMoreQuantifier
	| PLUS (reluctant = QUESTION_MARK)?								# oneOrMoreQuantifier
	| QUESTION_MARK (reluctant = QUESTION_MARK)?					# zeroOrOneQuantifier
	| '{' exactly = INTEGER_VALUE '}' (reluctant = QUESTION_MARK)?	# rangeQuantifier
	| '{' (atLeast = INTEGER_VALUE)? ',' (atMost = INTEGER_VALUE)? '}' (
		reluctant = QUESTION_MARK
	)? # rangeQuantifier;

updateAssignment: identifier EQ expression;

explainOption:
	FORMAT value = (TEXT | GRAPHVIZ | JSON)					# explainFormat
	| TYPE value = (LOGICAL | DISTRIBUTED | VALIDATE | IO)	# explainType;

transactionMode:
	ISOLATION LEVEL levelOfIsolation	# isolationLevel
	| READ accessMode = (ONLY | WRITE)	# transactionAccessMode;

levelOfIsolation:
	READ UNCOMMITTED	# readUncommitted
	| READ COMMITTED	# readCommitted
	| REPEATABLE READ	# repeatableRead
	| SERIALIZABLE		# serializable;

callArgument:
	expression						# positionalArgument
	| identifier '=>' expression	# namedArgument;

pathElement:
	identifier '.' identifier	# qualifiedArgument
	| identifier				# unqualifiedArgument;

pathSpecification: pathElement (',' pathElement)*;

privilege: SELECT | DELETE | INSERT | UPDATE;

qualifiedName: identifier ('.' identifier)*;

grantor:
	principal		# specifiedPrincipal
	| CURRENT_USER	# currentUserGrantor
	| CURRENT_ROLE	# currentRoleGrantor;

principal:
	identifier			# unspecifiedPrincipal
	| USER identifier	# userPrincipal
	| ROLE identifier	# rolePrincipal;

roles: identifier (',' identifier)*;

identifier
    : IDENTIFIER             #unquotedIdentifier
    | QUOTED_IDENTIFIER      #quotedIdentifier
    | nonReserved            #unquotedIdentifier
    | BACKQUOTED_IDENTIFIER  #backQuotedIdentifier
    | DIGIT_IDENTIFIER       #digitIdentifier
    ;

number:
	MINUS? DECIMAL_VALUE	# decimalLiteral
	| MINUS? DOUBLE_VALUE	# doubleLiteral
	| MINUS? INTEGER_VALUE	# integerLiteral;

nonReserved
	: ADD
	| ADMIN
	| AFTER
	| ALL
	| ANALYZE
	| ANY
	| ARRAY
	| ASC
	| AT
	| AUTHORIZATION
	| BERNOULLI
	| CALL
	| CASCADE
	| CATALOGS
	| COLUMN
	| COLUMNS
	| COMMENT
	| COMMIT
	| COMMITTED
	| CURRENT
	| DATA
	| DATE
	| DAY
	| DEFAULT
	| DEFINE
	| DEFINER
	| DESC
	| DISTRIBUTED
	| DOUBLE
	| EMPTY
	| EXCLUDING
	| EXPLAIN
	| FETCH
	| FILTER
	| FINAL
	| FIRST
	| FOLLOWING
	| FORMAT
	| FUNCTIONS
	| GRANT
	| GRANTED
	| GRANTS
	| DENY
	| GRAPHVIZ
	| GROUPS
	| HOUR
	| IF
	| IGNORE
	| INCLUDING
	| INITIAL
	| INPUT
	| INTERVAL
	| INVOKER
	| IO
	| ISOLATION
	| JSON
	| LAST
	| LATERAL
	| LEVEL
	| LIMIT
	| LOCAL
	| LOGICAL
	| MAP
	| MATCH
	| MATCHED
	| MATCHES
	| MATCH_RECOGNIZE
	| MATERIALIZED
	| MEASURES
	| MERGE
	| MINUTE
	| MONTH
	| NEXT
	| NFC
	| NFD
	| NFKC
	| NFKD
	| NO
	| NONE
	| NULLIF
	| NULLS
	| OFFSET
	| OMIT
	| ONE
	| ONLY
	| OPTION
	| ORDINALITY
	| OUTPUT
	| OVER
	| PARTITION
	| PARTITIONS
	| PAST
	| PATH
	| PATTERN
	| PER
	| PERMUTE
	| POSITION
	| PRECEDING
	| PRECISION
	| PRIVILEGES
	| PROPERTIES
	| RANGE
	| READ
	| REFRESH
	| RENAME
	| REPEATABLE
	| REPLACE
	| RESET
	| RESPECT
	| RESTRICT
	| REVOKE
	| ROLE
	| ROLES
	| ROLLBACK
	| ROW
	| ROWS
	| RUNNING
	| SCHEMA
	| SCHEMAS
	| SECOND
	| SECURITY
	| SEEK
	| SERIALIZABLE
	| SESSION
	| SET
	| SETS
	| SHOW
	| SOME
	| START
	| STATS
	| SUBSET
	| SUBSTRING
	| SYSTEM
	| TABLES
	| TABLESAMPLE
	| TEXT
	| TIES
	| TIME
	| TIMESTAMP
	| TO
	| TRANSACTION
	| TRUNCATE
	| TRY_CAST
	| TYPE
	| UNBOUNDED
	| UNCOMMITTED
	| UNMATCHED
	| UPDATE
	| USE
	| USER
	| VALIDATE
	| VERBOSE
	| VIEW
	| WINDOW
	| WITHOUT
	| WORK
	| WRITE
	| YEAR
	| ZONE;

ADD: 'ADD';
ADMIN: 'ADMIN';
AFTER: 'AFTER';
ALL: 'ALL';
ALTER: 'ALTER';
ANALYZE: 'ANALYZE';
AND: 'AND';
ANY: 'ANY';
ARRAY: 'ARRAY';
AS: 'AS';
ASC: 'ASC';
AT: 'AT';
AUTHORIZATION: 'AUTHORIZATION';
BERNOULLI: 'BERNOULLI';
BETWEEN: 'BETWEEN';
BY: 'BY';
CALL: 'CALL';
CASCADE: 'CASCADE';
CASE: 'CASE';
CAST: 'CAST';
CATALOGS: 'CATALOGS';
COLUMN: 'COLUMN';
COLUMNS: 'COLUMNS';
COMMENT: 'COMMENT';
COMMIT: 'COMMIT';
COMMITTED: 'COMMITTED';
CONSTRAINT: 'CONSTRAINT';
CREATE: 'CREATE';
CROSS: 'CROSS';
CUBE: 'CUBE';
CURRENT: 'CURRENT';
CURRENT_CATALOG: 'CURRENT_CATALOG';
CURRENT_DATE: 'CURRENT_DATE';
CURRENT_PATH: 'CURRENT_PATH';
CURRENT_ROLE: 'CURRENT_ROLE';
CURRENT_SCHEMA: 'CURRENT_SCHEMA';
CURRENT_TIME: 'CURRENT_TIME';
CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP';
CURRENT_USER: 'CURRENT_USER';
DATA: 'DATA';
DATE: 'DATE';
DAY: 'DAY';
DEFAULT: 'DEFAULT';
DEALLOCATE: 'DEALLOCATE';
DEFINER: 'DEFINER';
DELETE: 'DELETE';
DESC: 'DESC';
DESCRIBE: 'DESCRIBE';
DEFINE: 'DEFINE';
DISTINCT: 'DISTINCT';
DISTRIBUTED: 'DISTRIBUTED';
DOUBLE: 'DOUBLE';
DROP: 'DROP';
ELSE: 'ELSE';
EMPTY: 'EMPTY';
END: 'END';
ESCAPE: 'ESCAPE';
EXCEPT: 'EXCEPT';
EXCLUDING: 'EXCLUDING';
EXECUTE: 'EXECUTE';
EXISTS: 'EXISTS';
EXPLAIN: 'EXPLAIN';
EXTRACT: 'EXTRACT';
FALSE: 'FALSE';
FETCH: 'FETCH';
FILTER: 'FILTER';
FINAL: 'FINAL';
FIRST: 'FIRST';
FOLLOWING: 'FOLLOWING';
FOR: 'FOR';
FORMAT: 'FORMAT';
FROM: 'FROM';
FULL: 'FULL';
FUNCTIONS: 'FUNCTIONS';
GRANT: 'GRANT';
GRANTED: 'GRANTED';
GRANTS: 'GRANTS';
DENY: 'DENY';
GRAPHVIZ: 'GRAPHVIZ';
GROUP: 'GROUP';
GROUPING: 'GROUPING';
GROUPS: 'GROUPS';
HAVING: 'HAVING';
HOUR: 'HOUR';
IF: 'IF';
IGNORE: 'IGNORE';
IN: 'IN';
INCLUDING: 'INCLUDING';
INITIAL: 'INITIAL';
INNER: 'INNER';
INPUT: 'INPUT';
INSERT: 'INSERT';
INTERSECT: 'INTERSECT';
INTERVAL: 'INTERVAL';
INTO: 'INTO';
INVOKER: 'INVOKER';
IO: 'IO';
IS: 'IS';
ISOLATION: 'ISOLATION';
JOIN: 'JOIN';
JSON: 'JSON';
LAST: 'LAST';
LATERAL: 'LATERAL';
LEFT: 'LEFT';
LEVEL: 'LEVEL';
LIKE: 'LIKE';
LIMIT: 'LIMIT';
LOCAL: 'LOCAL';
LOCALTIME: 'LOCALTIME';
LOCALTIMESTAMP: 'LOCALTIMESTAMP';
LOGICAL: 'LOGICAL';
MAP: 'MAP';
MATCH: 'MATCH';
MATCHED: 'MATCHED';
MATCHES: 'MATCHES';
MATCH_RECOGNIZE: 'MATCH_RECOGNIZE';
MATERIALIZED: 'MATERIALIZED';
MEASURES: 'MEASURES';
MERGE: 'MERGE';
MINUTE: 'MINUTE';
MONTH: 'MONTH';
NATURAL: 'NATURAL';
NEXT: 'NEXT';
NFC: 'NFC';
NFD: 'NFD';
NFKC: 'NFKC';
NFKD: 'NFKD';
NO: 'NO';
NONE: 'NONE';
NORMALIZE: 'NORMALIZE';
NOT: 'NOT';
NULL: 'NULL';
NULLIF: 'NULLIF';
NULLS: 'NULLS';
OFFSET: 'OFFSET';
OMIT: 'OMIT';
ON: 'ON';
ONE: 'ONE';
ONLY: 'ONLY';
OPTION: 'OPTION';
OR: 'OR';
ORDER: 'ORDER';
ORDINALITY: 'ORDINALITY';
OUTER: 'OUTER';
OUTPUT: 'OUTPUT';
OVER: 'OVER';
PARTITION: 'PARTITION';
PARTITIONS: 'PARTITIONS';
PAST: 'PAST';
PATH: 'PATH';
PATTERN: 'PATTERN';
PER: 'PER';
PERMUTE: 'PERMUTE';
POSITION: 'POSITION';
PRECEDING: 'PRECEDING';
PRECISION: 'PRECISION';
PREPARE: 'PREPARE';
PRIVILEGES: 'PRIVILEGES';
PROPERTIES: 'PROPERTIES';
RANGE: 'RANGE';
READ: 'READ';
RECURSIVE: 'RECURSIVE';
REFRESH: 'REFRESH';
RENAME: 'RENAME';
REPEATABLE: 'REPEATABLE';
REPLACE: 'REPLACE';
RESET: 'RESET';
RESPECT: 'RESPECT';
RESTRICT: 'RESTRICT';
REVOKE: 'REVOKE';
RIGHT: 'RIGHT';
ROLE: 'ROLE';
ROLES: 'ROLES';
ROLLBACK: 'ROLLBACK';
ROLLUP: 'ROLLUP';
ROW: 'ROW';
ROWS: 'ROWS';
RUNNING: 'RUNNING';
SCHEMA: 'SCHEMA';
SCHEMAS: 'SCHEMAS';
SECOND: 'SECOND';
SECURITY: 'SECURITY';
SEEK: 'SEEK';
SELECT: 'SELECT';
SERIALIZABLE: 'SERIALIZABLE';
SESSION: 'SESSION';
SET: 'SET';
SETS: 'SETS';
SHOW: 'SHOW';
SOME: 'SOME';
START: 'START';
STATS: 'STATS';
SUBSET: 'SUBSET';
SUBSTRING: 'SUBSTRING';
SYSTEM: 'SYSTEM';
TABLE: 'TABLE';
TABLES: 'TABLES';
TABLESAMPLE: 'TABLESAMPLE';
TEXT: 'TEXT';
THEN: 'THEN';
TIES: 'TIES';
TIME: 'TIME';
TIMESTAMP: 'TIMESTAMP';
TO: 'TO';
TRANSACTION: 'TRANSACTION';
TRUNCATE: 'TRUNCATE';
TRUE: 'TRUE';
TRY_CAST: 'TRY_CAST';
TYPE: 'TYPE';
UESCAPE: 'UESCAPE';
UNBOUNDED: 'UNBOUNDED';
UNCOMMITTED: 'UNCOMMITTED';
UNION: 'UNION';
UNMATCHED: 'UNMATCHED';
UNNEST: 'UNNEST';
UPDATE: 'UPDATE';
USE: 'USE';
USER: 'USER';
USING: 'USING';
VALIDATE: 'VALIDATE';
VALUES: 'VALUES';
VERBOSE: 'VERBOSE';
VIEW: 'VIEW';
WHEN: 'WHEN';
WHERE: 'WHERE';
WINDOW: 'WINDOW';
WITH: 'WITH';
WITHOUT: 'WITHOUT';
WORK: 'WORK';
WRITE: 'WRITE';
YEAR: 'YEAR';
ZONE: 'ZONE';

EQ: '=';
NEQ: '<>' | '!=';
LT: '<';
LTE: '<=';
GT: '>';
GTE: '>=';

PLUS: '+';
MINUS: '-';
ASTERISK: '*';
SLASH: '/';
PERCENT: '%';
CONCAT: '||';
QUESTION_MARK: '?';

STRING: '\'' ( ~'\'' | '\'\'')* '\'';

UNICODE_STRING: 'U&\'' ( ~'\'' | '\'\'')* '\'';

// Note: we allow any character inside the binary literal and validate its a correct literal when
// the AST is being constructed. This allows us to provide more meaningful error messages to the
// user
BINARY_LITERAL: 'X\'' (~'\'')* '\'';

INTEGER_VALUE: DIGIT+;

DECIMAL_VALUE: DIGIT+ '.' DIGIT* | '.' DIGIT+;

DOUBLE_VALUE:
	DIGIT+ ('.' DIGIT*)? EXPONENT
	| '.' DIGIT+ EXPONENT;

IDENTIFIER: (LETTER | '_') (LETTER | DIGIT | '_')*;

DIGIT_IDENTIFIER: DIGIT (LETTER | DIGIT | '_')+;

QUOTED_IDENTIFIER: '"' ( ~'"' | '""')* '"';

BACKQUOTED_IDENTIFIER: '`' ( ~'`' | '``')* '`';
SEMICOLON: ';';

fragment EXPONENT: 'E' [+-]? DIGIT+;

fragment DIGIT: [0-9];

fragment LETTER: [A-Za-z];

SIMPLE_COMMENT: '--' ~[\r\n]* '\r'? '\n'? -> channel(HIDDEN);

BRACKETED_COMMENT: '/*' .*? '*/' -> channel(HIDDEN);

WS: [ \r\n\t]+ -> channel(HIDDEN);

// Catch-all for anything we can't recognize. We use this to be able to ignore and recover all the
// text when splitting statements with DelimiterLexer
UNRECOGNIZED: .;
